/*
  
  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/
  
  ===================================================
  
  Name:       CCM (bookKeeping.C)
  
  Copyright:  Damian J. J. Farnell
  
  Date:       Wed Sep 29 1999
  
  Purpose:    Determine and solve CCM LSUBm equations.

  Rewritten:  Joerg Schulenburg 2005
              - cout replaced by printf (was mixed with printf)
  
  ===================================================
  
  ---------------------------------------------------

  MODULE NAME: BOOKKEEPING.C 

  PURPOSE:     This module does all of the `bookkeeping' 
               for all of the other modules. Hence, all
               routines that are multiply used throughout
               the code are contained here. 

  ---------------------------------------------------

*/


// #include<iostream.h>
#include<stdlib.h>
#include<math.h>
#include<stdio.h>
#include<string.h>
#if (PARALLEL_FLAG>0)
#include<mpi.h>
#endif

#include"../include/nr.h"
#include"../include/nrutil.h"
#include"../include/Constants.h"
#include"../include/Externals.h"
#include"../include/FunctionPrototypes.h"

void printOutConfig(char *s1, int (*tempConfiguration)[5], int order) {
      int j;
      for(j=1;j<=order;j++) // (x y z uc)
	printf("%s [%d] (%d, %d, %d, %d)\n",
               s1, j,
	       tempConfiguration[j][1],
	       tempConfiguration[j][2],
	       tempConfiguration[j][3],
	       tempConfiguration[j][4]);
}

// storage place for (Params.s2+1)^(j-1) to speedup code
long int divisors[BLOCK_LENGTH+3];

void initDivisors(void) {
  int tempS, j;
  tempS = Params.s2+1;  // 2s+1
  divisors[1] = 1;
  for(j=2; j<=BLOCK_LENGTH+1; j++) {
    divisors[j] = divisors[j-1] * tempS;  // tempS ^ (j-1)
    // check for overflow, compare with product of doubles (Joerg 2005)
    if (divisors[j] * 1.0 != divisors[j-1] * 1.0 * tempS) {
      printf(" tempS^%d causes integer overflow\n", j-1);
      exit(1);
    }
  }
}


// in: order of ccm (maximum spinflips)
//     configuration[1...order?][5]
// out: descriptor[MAX_SPIN_LIST]
// cost most CPU time (Joerg Oct2004)
void findUniqueDescriptorOfCluster(int order,
				   int **configuration,
				   int *descriptor){

  /* This routine gets a set of crystallographic lattice sites
     which define a cluster and returns a set of unique integer 
     numbers in `descriptor' which fully describe it.

     Note that this is its FUNDAMENTAL description, i.e., 
     of the how ever many possible symmetry equivalent 
     cluster there are of it, this `descriptor' has 
     the lowest numerical value.

     PARTS ARE:

     1) determine symmetry related equivalents and translate
     ...into the x>=0,y>=0,z>=0 part.

     2) determine the `spin' form of these equivalents

     3) turn this `spin' form into a unique set of integers

     4) find set of integers with lowest `numerical' value
     ...out of all the symmetry related clusters and return
     ...this in `descriptor'.

  */

  int i,j,k,l;
  int minR[3];  // x y z
  int OK; 
  // int **tempConfiguration;  // obsolete
  static int tempDescriptor[MAX_SPIN_LIST];

  // tempConfiguration=matrix(0,MAX_LSUBM+1,0,5);  // to slow
  static int tempConfiguration[MAX_LSUBM+1][5];

  for(k=1;k<MAX_SPIN_LIST;k++)
    descriptor[k]=1000000000;
  
  if (Control.output >4)
      printf("findUniqueDescriptorOfCluster:\n");
  
  for(i=1;i<=LatticeSymmetries.number;i++) {

    // find symmetry related clusters

    symmetryGroup(i,order,configuration,tempConfiguration);

    if (Control.output >4)
      printOutConfig("transformed lattice site:",tempConfiguration, order);
    
    // translate into correct quadrant, x>=0,y>=0,z>=0
    
    for (k=0; k<3; k++) minR[k]=10000000;

    for(j=1;j<=order;j++)
      for (k=0; k<3; k++)
        if (tempConfiguration[j][k+1]<minR[k]) 
           minR[k] = tempConfiguration[j][k+1];

    for(j=1;j<=order;j++)
      for (k=0; k<3; k++)
        tempConfiguration[j][k+1] -= minR[k];

    if (Control.output >4)
      printOutConfig("translated  lattice site:",tempConfiguration, order);

    // now find the descriptor (costs most CPU, Joerg Oct04)
    
    findConfigurationToDescriptor(tempDescriptor, tempConfiguration, order);
    // if tempDescriptor[]=999999 we have a invalid cluster

    if (Control.output > 4) {
      printf("Temp. Descriptor(%d) = ",i); // i = number of lattice symmetry
      printOutDescriptor( tempDescriptor );
    }

    // check if tempDesc is smaller than Desc

    for (OK=0,k=(MAX_SPIN_LIST-1); k>0; k--) {
      if (tempDescriptor[k] > descriptor[k])  break;
      if (tempDescriptor[k] < descriptor[k]) { OK=1; break; }
    }

    if (OK) {
      for(k=1;k<MAX_SPIN_LIST;k++)
	descriptor[k]=tempDescriptor[k];
    }

  }

#if DEBUG
  if (descriptor[1]==1000000000) {
    printf("Can't find a reasonable descriptor of this cluster!\n"
      "Exiting...\n\n");
    exit(1);
  }
#endif

  if (Control.output > 6) {
    printf(" Minimum Descriptors = ");
    printOutDescriptor( descriptor );
  }
  // free_matrix(tempConfiguration,0,MAX_LSUBM+1,0,5);
 
}

// in:  configuration[1..order][1..4] 
// out: tempConfiguration[1..order][1..4] - symmetric coords 
void symmetryGroup(int switchVariable, 
		   int order,
		   int **configuration,
		   int (*tempConfiguration)[5])
{

  int i;

  double newx, newy, newz;
  double x,y,z;

  // symmorphic point group symetries entered here only =>
  // triangular lattice not-yet implemented, to be done.

  if (switchVariable > 48 || switchVariable < 1) {
    printf("Switch Variable out of range %d...\nExiting!\n\n", switchVariable);
    exit(1);
  }

  for(i=1;i<=order;i++) {
    
    findCartesiansFromGeneralisedPosition(x, y, z, configuration[i]);
    
    newx=x*LatticeSymmetries.R[switchVariable][1][1]+
         y*LatticeSymmetries.R[switchVariable][1][2]+
         z*LatticeSymmetries.R[switchVariable][1][3]+
           LatticeSymmetries.x[switchVariable];
    
    newy=x*LatticeSymmetries.R[switchVariable][2][1]+
         y*LatticeSymmetries.R[switchVariable][2][2]+
         z*LatticeSymmetries.R[switchVariable][2][3]+
           LatticeSymmetries.y[switchVariable];
    
    newz=x*LatticeSymmetries.R[switchVariable][3][1]+
         y*LatticeSymmetries.R[switchVariable][3][2]+
         z*LatticeSymmetries.R[switchVariable][3][3]+
           LatticeSymmetries.z[switchVariable];
  
    // ToDo: this consumes most of the time Sep05, reduce for 1D ?  
    findGeneralisedPositionFromCartesians(newx,newy,newz,tempConfiguration[i]);
  }
}

// in:  x y z
// out: configuration[1..4]
// ToDo: called very often by obtain! takes much cpu time, Joerg Sep05
void findGeneralisedPositionFromCartesians(double x, 
					   double y, 
					   double z, 
					   int *configuration)
{
  static int index[5];
  int i,j,k,l,found;

  // double **m, **tempm;
  // double *s;
  double distance;
  // int *ind;

  double d;

  static int  *ind=ivector(0,5);    // call it only once for faster code (Joerg)
  static double *s=dvector(0,5);
  static double **m=dmatrix(0,5,0,5);      // slow
  static double **tempm=dmatrix(0,5,0,5);  // slow
  static int n_out = 0; // limit the output
  static double non_ortho=0;

  // this is a rather complicated way of doing this though quite
  // quick -- see linear search below.
  
  // for simpler lattices this whole piece of code can be replaced
  // by a far simpler algroithm.
  
  // note that this code is implemented in 2D only here. ???
  
  static int do_once=0;
  if (!do_once){
    for (k=1; k<=3; k++) {
      m[k][1]=BravaisLattice.a[k];
      m[k][2]=BravaisLattice.b[k];
      m[k][3]=BravaisLattice.c[k];
      non_ortho+=fabs(m[k][1]*m[k][2])
                +fabs(m[k][2]*m[k][3])
                +fabs(m[k][3]*m[k][1]);
    }
    if (non_ortho)
      printf("BravaisLattice is not orthogonal, PG is slower\n");
    do_once=1;
  }

  found=0;

  index[4]=-1;
  for(i=1;i<=UnitCell.number;i++) {

    s[1]=x-UnitCell.x[i];
    s[2]=y-UnitCell.y[i];
    s[3]=z-UnitCell.z[i];

#if 1 // This code seems to unneccessary, but cost most of the time (Joerg)
      // only needed if Bravais is non-orthogonal (ToDo: check that)
    if (non_ortho) { // Joerg Sep05, SpeedUp for dim=1!
      // tempm is work space and will be destroyed
      for(j=1;j<=3;j++)
        for(k=1;k<=3;k++)
  	  tempm[j][k]=m[j][k];

      // what happens here? Joerg
      // input s[3],
      //       tempm=Bravais+workspace (just 1 for LC)
      ludcmp(tempm,3,ind,&d); // ToDo: most cpu, replace or make Dim-dependent
      lubksb(tempm,3,ind,s);
      // input + output is s[3]

      if (n_out<10)  // just to see if ludcmp call is really needed (1D=no)
      if (fabs(s[1]-(x-UnitCell.x[i]))
         +fabs(s[2]-(y-UnitCell.y[i]))
         +fabs(s[3]-(z-UnitCell.z[i]))>0.0) { n_out++;
        printf("findGeneralisedPositionFromCartesians i=%d n_out=%d:\n",i,n_out);
        printf("before luXX: ");
        printf(" %9.5f",x-UnitCell.x[i]);
        printf(" %9.5f",y-UnitCell.y[i]);
        printf(" %9.5f",z-UnitCell.z[i]);
        printf("\n");
        printf("after  luXX: ");
        printf(" %9.5f",s[1]);
        printf(" %9.5f",s[2]);
        printf(" %9.5f",s[3]);
        printf("\n");
      }
    }
#endif

    // check if the values for s[] are integers.

    for (k=1; k<=3; k++) {
      index[k] = ((s[k]<0)?-1:+1) * (int) (fabs(s[k])+0.5);
    }

    index[4]=i;

    distance =fabs(x-index[1]*BravaisLattice.a[1]
                    -index[2]*BravaisLattice.b[1]
	 	    -index[3]*BravaisLattice.c[1]-UnitCell.x[index[4]]);
    distance+=fabs(y-index[1]*BravaisLattice.a[2]
                    -index[2]*BravaisLattice.b[2]
		    -index[3]*BravaisLattice.c[2]-UnitCell.y[index[4]]);
    distance+=fabs(z-index[1]*BravaisLattice.a[3]
                    -index[2]*BravaisLattice.b[3]
		    -index[3]*BravaisLattice.c[3]-UnitCell.z[index[4]]);
    if (distance < 0.001) {   // if integers
      found=1;
      break;
    }
  }  // i = 1...UnitCell

  if (!found) { 
    printf("Error in you findGeneralisedPositionFromCartesians %d %d %d %d\n"
           "%d %d %d\n", index[1], index[2], index[3], index[4], x, y, z);
    exit(0);
  }

//  free_dmatrix(m,0,5,0,5);
//  free_dmatrix(tempm,0,5,0,5);
//  free_ivector(ind,0,5);
//  free_dvector(s,0,5);

  configuration[1]=index[1];
  configuration[2]=index[2];
  configuration[3]=index[3];
  configuration[4]=index[4];
}


// caled by bookKeeping.C and define.C
void findCartesiansFromGeneralisedPosition(double &x, 
					   double &y, 
					   double &z, 
					   int *configuration)
{

  // this routine finds the cartesian position from the 
  // `generalised' lattice position => bravais vector + primitive
  // cell vector

  x=BravaisLattice.a[1]*configuration[1]+
    BravaisLattice.b[1]*configuration[2]+
    BravaisLattice.c[1]*configuration[3]+
            UnitCell.x[configuration[4]];

  y=BravaisLattice.a[2]*configuration[1]+
    BravaisLattice.b[2]*configuration[2]+
    BravaisLattice.c[2]*configuration[3]+
             UnitCell.y[configuration[4]];

  z=BravaisLattice.a[3]*configuration[1]+
    BravaisLattice.b[3]*configuration[2]+
    BravaisLattice.c[3]*configuration[3]+
             UnitCell.z[configuration[4]];

}


// coding the configuration into a descriptor index (splitted to integers)
// out: descriptor[1..MAX_SPIN_LIST]=...,999999(=invalid)
// in:  config[1..order][1..4]=0..btLength for [1..3], ??? for [4]
// called by findUniqueDescriptorOfCluster() in bookKeeping.C
// not a good name I think (Joerg05)
void findConfigurationToDescriptor(int *descriptor,
				   int (*config)[5],
				   int order)
{
  int i,j,k,l,a,sum_abs;
  int index;
  int tempNum;
  int block;

  for(i=1;i<MAX_SPIN_LIST;i++) descriptor[i]=0;  // reset descriptor bits
  
  // the first part is to determine a unique lattice index in the bounding
  // triangle for each spin flip operator in a given cluster. Each spin
  // flip is denoted by it bravais lattice vector + its number in the 
  // unit cell. 
  
  // each cell in the the bounding triangle is a unit cell.

  for(i=1;i<=order;i++) {

    // check if within bounding triangle in 1D, 2D and 3D
    for (sum_abs=0, j=1; j<=Params.dim; j++) sum_abs += abs(config[i][j]);
    if (sum_abs > Params.btLength-1) {
      // point lies outside the bounding triangle.
      for(j=1;j<MAX_SPIN_LIST;j++) descriptor[j]=999999;  /* Sep05 */
      return;
    }

    // get a unique index for every possible site
    if (Params.dim == 1) {
      index = config[i][1] + 1;
    } else if (Params.dim == 2) {
      index = BTarray[ config[i][1] ][ config[i][2] ][0];
    } else {
      index = BTarray[ config[i][1] ][ config[i][2] ][ config[i][3] ];
    }   

    // at this point: 1 <= index <= max_numbers_cells_in_BT

    // we now have an unique site ``index'' for 1D-3D which completely
    // specifies the position of a given spin in the bounding triangle.
    // we now wish to encode this information in a given ``block'' 
    // with a given number ``tempNum'' in ``descriptor''.

    // note that BLOCK_LENGTH gives the numbers of ``spins'' the
    // each element of the descriptor describes. Normally set to
    // 20 such spins so that we can look at large values of the 
    // quantum spin number.

    index = (index-1)*UnitCell.number + config[i][4] - 1;

    // at this point: 1 <= index <= max_number_cells_BT*UnitCell.number
    // note that: 0 <= config[][4] < UnitCell.number

    block  = index / BLOCK_LENGTH;       // 0 .. MAX_SPIN_LIST-2

    index  = index % BLOCK_LENGTH;       // 0 .. BLOCK_LENGTH-1
    
    tempNum = divisors[index+1]; // (Params.s2+1)^(i-1)=(2s+1)^(i-1)

    if (block+1 >= MAX_SPIN_LIST) { // (max_BTarray*UC)/BLOCK_LENGTH
       printf("Error: add 1 to MAX_SPIN_LIST\n"); exit(1);
    }
    descriptor[block+1] += tempNum;
  }  

}

void printOutDescriptor(int *descriptor){
  int i;
  printf("%d number descriptor:  ",MAX_SPIN_LIST-1);
  for(i=1;i<MAX_SPIN_LIST;i++) 
    printf("%d ",descriptor[i]);
  printf("\n");

}

// almost the same as findClusterFromDescriptor (Joerg)
void printOutClusterFromDescriptor(int *descriptor){

  int i,j,k,l,m;
  int index;
  int temp1, temp2, temp, tempS;
//  long int divisors[BLOCK_LENGTH+3];

  int x,y,z;

  int cell;
  static int checkBlock[400][MAX_UNIT_CELL];
  
  int flips, count;
  
  // checkBlock makes sure that no more than 2*Params.m+1 spin
  // flips per site occur.
  
  for(i=0;i<=max_number_cells_BT;i++)
    for(j=0;j<=UnitCell.number;j++)
      checkBlock[i][j]=0;

  // now find the divisors which give the numbers to put into checkBlock

#if 0
  tempS=Params.s2+1; // 2s+1
  divisors[1]=1;
  for(j=2;j<=BLOCK_LENGTH+1;j++) {
    divisors[j]=divisors[j-1]*tempS;
  }
#endif

  for(i=1;i<MAX_SPIN_LIST;i++) {    
    if (descriptor[i] > 0) {
      temp1=descriptor[i];
      for(j=BLOCK_LENGTH;j>0;j--) {
	
	temp2=(int) (temp1/divisors[j]);     
	temp1=temp1-temp2*divisors[j];  
	
	index=j+(i-1)*BLOCK_LENGTH;

	// note that: 1 <= index < BLOCK_LENGTH*MAX_SPIN_LIST

	temp=index;
	index=(int) ((index-1)/UnitCell.number);
	cell=temp-UnitCell.number*index;
	index++;

	// 1 <= index <= (BLOCK_LENGTH*MAX_SPIN_LIST-1)/UnitCell.number 

	if (index > max_number_cells_BT && temp2 > 0) {
	  printf("index (=%d) > max_number_cells_BT (=%d)...\nExiting...\n\n",
                index, max_number_cells_BT);
	  exit(1);
	} 

	if (temp2 > 0)
	  checkBlock[index][cell] += temp2;

	if (temp2 > 0 && Control.output > 4)
	  printf("j = %d index = %d cell = %d temp2 = %d\n",
                  j, index, cell, temp2);
	
      }
    }
  }

  // now count the number of spin flips

  flips=0;

  for(i=1;i<=max_number_cells_BT;i++) {
    for(j=1;j<=UnitCell.number;j++) {

      if (checkBlock[i][j] > 0) {
	  x=lookUpTable[i][1];
	  y=lookUpTable[i][2];
	  z=lookUpTable[i][3];
      }
      
      for (m=1;m<=checkBlock[i][j];m++) {
	flips++;
	printf("lattice site(%d): (%d, %d, %d, %d)\n",flips,x,y,z,j);
      }
    }
  }
  printf("\n");
}


void printOutCluster(int length, int **configuration)
{	
  int l; 
  
  for(l=1;l<=length;l++) {   // x  y   z   UC.number
    printf("lattice site(%d): (%d, %d, %d, %d)\n",
	   l,
	   configuration[l][1],
	   configuration[l][2],
	   configuration[l][3],
	   configuration[l][4]);
  }
}

/*
 *  input:  descriptor[MAX_SPIN_LIST]=???
 *  output: configuration[flips][x,y,z,UnitCell]   commented by Joerg 2005
 *  called by: obtain, define via findFundamentalSpinConfigurations, ...
 */
int findClusterFromDescriptor(int *descriptor, int **configuration){
  int i,j,k,l,m;
  int index;
  int temp1, temp2, temp, tempS;
//  long int divisors[BLOCK_LENGTH+3];

  int x,y,z;

  int cell;
  static int checkBlock[400][MAX_UNIT_CELL];  // ToDo: replace 400, Joerg
  
  int flips, count;
  
  // checkBlock makes sure that no more than 2*Params.m+1 spin
  // flips per site occur.
  
  for(i=0;i<=max_number_cells_BT;i++)
    for(j=0;j<=UnitCell.number;j++)
      checkBlock[i][j]=0;

  // now find the divisors which give the numbers to put into checkBlock

#if 0  // done on input.C
  tempS=Params.s2+1;  // 2s+1
  divisors[1]=1;
  for(j=2;j<=BLOCK_LENGTH+1;j++) {
    divisors[j] = divisors[j-1] * tempS;  // tempS ^ (j-1)
    // check for overflow, compare with product of doubles (Joerg 2005)
    if (divisors[j] * 1.0 != divisors[j-1] * 1.0 * tempS) {
      printf(" tempS^%d causes integer overflow\n",j-1);
      exit(1);
    }
  }
#endif
  

  for(i=1;i<MAX_SPIN_LIST;i++) {    // default: 60
      if (descriptor[i] <= 0) continue;
      temp1=descriptor[i];
      for(j=BLOCK_LENGTH;j>0;j--) {      // default: 10
	
	temp2=(int) (temp1/divisors[j]);     // a/b
	temp1=temp1-temp2*divisors[j];       // a%b
	
	index=j+(i-1)*BLOCK_LENGTH;

	// note that: 1 <= index < BLOCK_LENGTH*MAX_SPIN_LIST

	temp=index;
	index=(int) ((index-1)/UnitCell.number);  // a/b
	cell=temp-UnitCell.number*index;          // a%b
	index++;

	// 1 <= index <= (BLOCK_LENGTH*MAX_SPIN_LIST-1)/UnitCell.number 

	if ((index > max_number_cells_BT || index>=400) && temp2 > 0) {
	  printf("index (=%d) > max_number_cells_BT (=%d)...\n" 
	         "Exiting...\n\n", index, max_number_cells_BT);
	  exit(1);
	} 

	if (temp2 > 0)
	  checkBlock[index][cell] += temp2;

	if (temp2 > 0 && Control.output > 4)
	  printf("j = %d index = %d cell = %d temp2 = %d\n",
                  j, index, cell, temp2) ;
	
      }
  }

  // now count the number of spin flips

  flips=0;

  for(i=1;i<=max_number_cells_BT;i++) {
    for(j=1;j<=UnitCell.number;j++) {

#if DEBUG
      if (checkBlock[i][j] > Params.s2){ 
	printf("Configuration has more than Params.s2 + 1 spin flips..."
	       "Exiting!\n\n");
	exit(1);
      }
#endif

      if (checkBlock[i][j] > 0) {
	  x=lookUpTable[i][1];   // see definePointsInBT() input.C
	  y=lookUpTable[i][2];
	  z=lookUpTable[i][3];
      }
      
      for (m=1;m<=checkBlock[i][j];m++) {  // checkBlock: 0 ... Params.s2
	flips++;
	configuration[flips][1]=x;
	configuration[flips][2]=y;
	configuration[flips][3]=z;
	configuration[flips][4]=j;	  
      }
    }
  }
  return flips;
}



// in: descriptor[1..(MAX_SPIN_LIST-1)] = spinconfig of 20 sites
// out: index to spinConfigurations with same descriptor
int findDescriptorIndex(int *descriptor,
			int numberElements) {

  int i,j;
  int t1, t2; // upper and lower border for binary search
  int tempPosition2;
  int middle;

  if (numberElements < 1) return 0;

#if 0 // dont happens
  for(i=1;i<MAX_SPIN_LIST;i++)
    if (descriptor[i]!=999999) break; // check for valid number != 999999
  if (i>=MAX_SPIN_LIST) return 0;     // all numbers are 999999
#endif

  // use a binary search and return the `position' of the element (if found!)
  // if NOT found then return zero.

  t1=1-1;
  t2=numberElements+1;

  // now check that elements in the `middle' do not match `descriptor'
  // Joerg Sep 05: lot of zeros are checked, replace MAX_SPIN_LIST by
  //               a calculated maximum value

  do {
    middle=(t1+t2)/2; // if t1+t2 is odd 
    for(i=MAX_SPIN_LIST-1;i>0;i--) {
      if (spinConfigurations[middle].descriptor[i]!=descriptor[i]) {
        if (spinConfigurations[middle].descriptor[i] > descriptor[i]) 
          t2=middle;
        else
          t1=middle;
        break;
      }
    }      
    if (i==0) return middle; 
  } while (t2-t1 > 1);
		       
  // old linear search routine. Useful to keep this as a check of 
  // new binary routine  

  /*
   for(j=numberElements;j>0;j--) {
     for(i=1;i<MAX_SPIN_LIST;i++)
       if (spinConfigurations[j].descriptor[i]!=descriptor[i]) break;
     if (i==MAX_SPIN_LIST) return j;
   } 
  */
  return 0; // not found
}

/* make binary bitmap spin[]={0..1,0..1,...} from index (Joerg) */
void binary(int index, int length, int *spin)
{
  int i, y, dummy=index;
  if (sizeof(int)*8 < length) { printf("binary Error OVERFLOW"); exit(1); }
  // for(i=1;i<=length;i++,index>>=1) { spin[i] = (index & 1); }
  for(i=1;i<=length;i++) {
    y= dummy/2;
    spin[i]= dummy-2*y;
    dummy=y;
  }
}

/* decode trimary code spin[]={0..2,0..2,...} from index1..3 (Joerg) */
void trimary(int index1, int index2, int index3, int length, int *spin)
{
  int i,y,dummy1, dummy2, dummy3;    
  dummy1=index1; /* spins 21..30 */
  dummy2=index2; /* spins 11..20 */
  dummy3=index3; /* spins  1..10 */
  if (length>30) { printf("trimary Error OVERFLOW"); exit(1); }

  for(i=21;i<=30 && i<=length;i++) {
    y= dummy1/3;
    spin[i]=dummy1-3*y;
    dummy1=y;
  }

  for(i=11;i<=20 && i<=length;i++) {
    y= dummy2/3;
    spin[i]=dummy2-3*y;
    dummy2=y;
  }

  for(i=1;i<=10 && i<=length;i++) {
    y= dummy3/3;
    spin[i]=dummy3-3*y;
    dummy3=y;
  }
}

/* make trimary bitmap spin[]={0..3,0..3,...} from index1..3 (Joerg) */
void quartary(int index1, int index2, int index3, int length, int *spin)
{
  
  int i,y,dummy1, dummy2, dummy3;    
  dummy1=index1;
  dummy2=index2;
  dummy3=index3;

  if (length>30) { printf("quartary Error OVERFLOW"); exit(1); }
    
  for(i=21;i<=30 && i<=length;i++) {
    y= dummy1/4;
    spin[i]=dummy1-4*y;
    dummy1=y;
  }

  for(i=11;i<=20 && i<=length;i++) {
    y= dummy2/4;
    spin[i]=dummy2-4*y;
    dummy2=y;
  }

  for(i=1;i<=10 && i<=length;i++) {
    y= dummy3/4;
    spin[i]=dummy3-4*y;
    dummy3=y;
  }
}

int findDescriptorIndex2(int *descriptor,
                        int numberElements) {

  int tempLogicFlag;
  int position;
  int i,j;
  int tempPosition1;
  int tempPosition2;
  int middle;
  int switchFlag;

  if (numberElements < 1)
    return 0;

  tempLogicFlag=1;

  // Joerg: do not understand that
  // if (descriptor[1]==999999) return 0; // not the same!

  for(i=1;i<MAX_SPIN_LIST;i++) {
    if (descriptor[i]!=999999) {
      tempLogicFlag=0;
      break;
    }
  }

  if (tempLogicFlag)
    return 0;

  position=0;

  // use a binary search and return the `position' of the element (if found!)
  // if NOT found then return zero.

  tempPosition1=1;
  tempPosition2=numberElements;

  // first check that the end elements do not match  `descriptor'

  tempLogicFlag=0;

  for(i=1;i<MAX_SPIN_LIST;i++) {
    if (latticeAnimals[tempPosition1].descriptor[i]!=descriptor[i]) {
      tempLogicFlag=1;
      break;
    }
  }
  
  if (!tempLogicFlag)
    position=tempPosition1;
  
  if (tempLogicFlag) {
    tempLogicFlag=0;
    for(i=1;i<MAX_SPIN_LIST;i++) {
      if (latticeAnimals[tempPosition2].descriptor[i]!=descriptor[i]) {
        tempLogicFlag=1;
        break;
      }
    }
    if (!tempLogicFlag)
      position=tempPosition2;
  }

  // now check that elements in the `middle' do not match `descriptor'

  if (tempLogicFlag) {
    do {

      middle=(tempPosition1+tempPosition2)/2;      
      tempLogicFlag=0;
      for(i=MAX_SPIN_LIST-1;i>0;i--) {
        if (latticeAnimals[middle].descriptor[i]!=descriptor[i]) {
          if (latticeAnimals[middle].descriptor[i] > descriptor[i])
            switchFlag=1;
          else
            switchFlag=0;
          tempLogicFlag=1;
          break;
        }
      }      

      if (!tempLogicFlag) {
        position=middle;
        tempPosition2=tempPosition1;
      } else {
        if (switchFlag)
          tempPosition2=middle;
        else
          tempPosition1=middle;
      }

    } while ((tempPosition2-tempPosition1) > 1);
  }
                       
  // old linear search routine. Useful to keep this as a check of 
  // new binary routine  

  /*
   for(j=numberElements;j>0;j--) {

    tempLogicFlag=0;
    
    for(i=1;i<MAX_SPIN_LIST;i++) {
      if (latticeAnimals[j].descriptor[i]!=descriptor[i]) {
        tempLogicFlag=1;
        break;
      }
    }

    if (!tempLogicFlag) {
      position=j;
      break;
    }
    } 
  */

  return position;

}




